#include <vector>
#include <string>

#include "Fab4.h"
#include "../name/panitz/util/Show.h"

using namespace std;
using namespace name::panitz::util;

namespace fab4 {

Var::Var(string n):name(n){};
string Var::toString(){return name;}

Constr::Constr(string n,vector<Expr*>* a):name(n),args(a){};
string Constr::toString(){return name+" "+show(args);}

App::App(Expr* p1,Expr* p2):p1(p1),p2(p2){} 
string App::toString(){return "("+show(p1)+" "+show(p2)+")";}

void Expr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void Var::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void Constr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void App::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void CaseExpr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void NumExpr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void CharExpr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}
void OpExpr::visit(ExprVisitor* vis){
  ((ExprVisitor*)vis)->eval(this);}

OpExpr::OpExpr(Expr* e1,Operator* op,Expr* e2):op(op){
    this->op=op;this->e1=e1;this->e2=e2;};
string OpExpr::toString(){return "("+show(e1)+show(op)+show(e2)+")";}

NumExpr::NumExpr(int n):n(n){};
string NumExpr::toString(){return show(n);}

CharExpr::CharExpr(char c):c(c){};
string CharExpr::toString(){return show(c);}

FunDef::FunDef(string name,vector<string> args,Expr* expr):
  name(name),args(args),expr(expr){}

FunDef::~FunDef(){
  delete expr;}

string FunDef::toString(){
  string result = name;
  for (int i=0;i<args.size();i=i+1)
    result = result+" "+args[i]; 
  result = result+" = "; 
  result = result+show(expr); 
  return result;
}

CaseAlt::CaseAlt(string constr,vector<string> vars,Expr* alt):
    constr(constr),vars(vars),alt(alt){};
string CaseAlt::toString(){
  return show(constr)+" "+show(vars)+" "+show(alt);}

CaseExpr::CaseExpr(Expr* e1,vector<CaseAlt*>* alts):
  e1(e1),alts(alts){};
string CaseExpr::toString(){
  return "case "+show(e1)+ " of "+show(alts);}

ConstrDef::ConstrDef(string name,int arity):name(name),arity(arity){}
std::string ConstrDef::toString(){
  return "constr "+name+" "+show(arity);
}

Fab4::Fab4(vector<ConstrDef*> cs,vector<FunDef*> fs,string n):
  constructors(cs),functions(fs),name(n){}

Fab4::~Fab4(){
  for (int i = 0;i<constructors.size();i=i+1)
    delete constructors[i];
  for (int i = 0;i<functions.size();i=i+1)
    delete functions[i];
}

std::string Fab4::toString(){
  string result="";
  for (int i = 0;i<constructors.size();i=i+1){
    if (i>0) result=result+";\n\n";
    result=result+show(constructors[i]);
  }
  for (int i = 0;i<functions.size();i=i+1){
    if (i>0||constructors.size()>0) result=result+";\n\n";
    result=result+show(functions[i]);
  }
          
  return result;
}
}


