#ifndef __FAB4_H
#define __FAB4_H 

#include <vector>
#include <string>
#include "../name/panitz/util/Visitor.h"
#include "Fab4Token.h"

using namespace std;
using namespace name::panitz::util;

namespace fab4 {

class Expr;
class ExprVisitor;

class Expr{
  public:
    virtual std::string toString()=0;
    virtual void visit(ExprVisitor* vis)=0;
};

class Var:public Expr{
  public:
    string name;
    Var(string n);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class CharExpr:public Expr{
  public:
    char c;
    CharExpr(char c);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class NumExpr:public Expr{
  public:
    int n;
    NumExpr(int n);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class OpExpr:public Expr{
  public:
    Expr* e1;
    Expr* e2;
    Operator* op;
    OpExpr(Expr* e1,Operator* op,Expr* e2);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class App:public Expr{
  public:
    Expr* p1;
    Expr* p2;
    App(Expr* p1,Expr* p2);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class Constr:public Expr{
  public:
    string name;
    vector<Expr*>* args;
    Constr(string n,vector<Expr*>* a);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class CaseAlt{
  public:
    string constr;
    vector<string> vars;
    Expr* alt;
    CaseAlt(string constr,vector<string> vars,Expr* alt);
    virtual string toString();
};

class CaseExpr:public Expr{
  public:
    Expr* e1;
    vector<CaseAlt*>* alts;
    CaseExpr(Expr* e1,vector<CaseAlt*>* alts);
    virtual string toString();
    virtual void visit(ExprVisitor* vis);
};

class ExprVisitor:public GeneralVisitor<Expr*>{
 public:
  virtual void eval(Expr* arg)=0;
  virtual void eval(Var* arg)=0;
  virtual void eval(Constr* arg)=0;
  virtual void eval(App* arg)=0;
  virtual void eval(CaseExpr* arg)=0;
  virtual void eval(NumExpr* arg)=0;
  virtual void eval(CharExpr* arg)=0;
  virtual void eval(OpExpr* arg)=0;
};

class FunDef{
  public:
   string name;
   vector<string> args;
   Expr* expr;
   FunDef(string name,vector<string> args,Expr* expr);

   virtual ~FunDef();
   virtual std::string toString();
};

class ConstrDef{
  public:
   string name;
   int arity;
   ConstrDef(string name,int arity);
   virtual std::string toString();
};

class Fab4 {
  public:
    string name;
    vector<ConstrDef*> constructors;
    vector<FunDef*> functions;

    Fab4(vector<ConstrDef*> cs,vector<FunDef*> fs,string n);
    virtual ~Fab4();
    virtual std::string toString();
};
}
#endif


