%option noyywrap

%{
#include <vector>
#include <string>
#include <sstream>
#include "../name/panitz/parser/ParsLib.h"
#include "../name/panitz/util/Show.h"
#include "Fab4Token.h"
%}

     using namespace std;
     using namespace fab4;
     int num_lines = 0, num_chars = 0;
     vector<Token*> result;
     string strLit;

CHAR       [\'].[\']
DIGIT      [0-9]
NAME       [a-z][a-zA-Z0-9]*
CNAME      [A-Z][a-zA-Z0-9]*
QUOTE      [\"]
WHITE      [\n\t\ ]

%x str
%%
\"        {strLit="";} BEGIN(str);

<str>\"   { /* saw closing quote - all done */
            BEGIN(INITIAL);
            result.insert(result.end(),new StringLiteral(strLit));
          }

<str>\n   {
            /* error - unterminated string constant */
            /* generate error message */
            throw "unterminated String constant";
          }

<str>\\[0-7]{1,3} {
            /* octal escape sequence */
             int result;
             (void) sscanf( yytext + 1, "%o", &result );
             if ( result > 0xff )
                throw "error, constant is out-of-bounds";
             strLit=strLit+show(result);
          }

<str>\\[0-9]+ {
             /* generate error - bad escape sequence; something
              * like '\48' or '\0777777'
              */
            throw "bad escape sequence";
          }

<str>\\n  strLit=strLit+ '\n';
<str>\\t  strLit=strLit+ '\t';
<str>\\r  strLit=strLit+ '\r';
<str>\\b  strLit=strLit+ '\b';
<str>\\f  strLit=strLit+ '\f';

<str>\\(.|\n)  strLit=strLit+ yytext[1];

<str>[^\\\n\"]+        {
       char *yptr = yytext;

       while ( *yptr ){strLit=strLit+ (*yptr++);}
     }

"constr"  result.insert(result.end(),CONSTR_T);
"case"    result.insert(result.end(),CASE_T);
"of"      result.insert(result.end(),OF_T);
"->"      result.insert(result.end(),ARROW_T);
";"       result.insert(result.end(),SEMICOLON_T);
"("       result.insert(result.end(),LPAR_T);
")"       result.insert(result.end(),RPAR_T);
"+"       result.insert(result.end(),ADD_T);
"-"       result.insert(result.end(),SUB_T);
"*"       result.insert(result.end(),MULT_T);
"/"       result.insert(result.end(),DIV_T);
"<="      result.insert(result.end(),LE_T);
">="      result.insert(result.end(),GE_T);
"<"       result.insert(result.end(),LT_T);
">"       result.insert(result.end(),GT_T);
"="       result.insert(result.end(),EQ_T);

{WHITE}   ;
{CHAR}    result.insert(result.end(),new CharLiteral(yytext[1] ));
{DIGIT}+  result.insert(result.end(),new Number(atoi( yytext ) ));
{NAME}    {result.insert(result.end(),new Name(yytext ) );}
{CNAME}   {result.insert(result.end(),new CName(yytext ) );}

%%
  namespace fab4 {

     vector<Token*> tokenize(string fileName){
       char fi [fileName.size()];
       fi [fileName.size()]='\0';
       fileName.copy(fi,fileName.size());
       yyin = fopen(fi, "r" );
       yylex();

       return result;}
     }


