#include "../name/panitz/parser/ParsLib.h"
#include "Fab4MkFunctions.h"

using namespace name::panitz::parser;
using namespace std;

namespace fab4 {
Parser<Token*,Token>* pTok(Token* t){
  return getToken(t,tokenEq);
}

Parser<Operator*,Token>* pOp(Token* t){
  return ((Parser<Operator*,Token>*)getToken(t,tokenEq));
}

Parser<Number*,Token>* pNum(Token* t){
  return ((Parser<Number*,Token>*)getToken(t,tokenEq));
}


Expr* mkNumExpr(Number* num){return new NumExpr(num->val);}
Expr* mkCharExpr(CharLiteral* cl){return new CharExpr(cl->c);}
Expr* mkStringExpr(StringLiteral* sl){
  vector<Expr*>* args=new vector<Expr*>();
  Expr* result=new Constr("Nil",args);
  for (int i=sl->name.size()-1;i>=0;i--){
      args=new vector<Expr*>();
    args->insert(args->end(),new CharExpr(sl->name[i]));
    args->insert(args->end(),result);
    result=new Constr("Cons",args);
  } 
  return result;
}

Expr* mkVar(Token* t){return new Var(t->name);}

Expr* mkOpExpr(Pair<Expr*,vector<Pair<Operator*,Expr*>*>*>* p){
  Expr* result=p->fst;
  vector<Pair<Operator*,Expr*>*>* opEs = p->snd;
  for (int i=0;i<opEs->size();i++){
    Operator* op=(*opEs)[i]->fst;
    Expr* e2 = (*opEs)[i]->snd;
    result = new OpExpr(result,op,e2);
  }
  return result;
}

Expr* mkFunApps(vector<Expr*>* exprs){
  Expr* result = (*exprs)[0];
  for (int i=1;i<exprs->size();i++)
    result=new App(result,(*exprs)[i]);
  delete exprs;
  return result;
}

Expr* mkConstr(Pair<Token*,vector<Expr*>*>* p){
  return new Constr(p->fst->name,p->snd);}

CaseAlt* mkCaseAlt(caseAltResult caseAlt){
  string cname = show(caseAlt->fst->fst->fst);
  vector<Token*>* varToks = caseAlt->fst->fst->snd;
  Expr* expr = caseAlt->snd;

  vector<string> vars;
  for (int i=0;i<varToks->size();i++)
      vars.insert(vars.end(),show((*varToks)[i]));

  delete caseAlt->fst->fst;
  delete caseAlt->fst;
  delete caseAlt;
  CaseAlt* result = new CaseAlt(cname,vars,expr);
  return result;
}

Expr* mkCaseExpr(caseParsResult casePars){
  Expr* e = casePars->fst->fst->snd;
  vector<CaseAlt*>* alts = casePars->snd;
  delete casePars->fst->fst;
  delete casePars->fst;
  delete casePars;
  return new CaseExpr(e,alts);
}

FunDef* mkFunDef
     (Pair<Pair<Pair<Token*,vector<Token*>*>*,Operator*>*,Expr*>* p){
  Token* name = p->fst->fst->fst;
  vector<Token*>* varTs = p->fst->fst->snd;
  vector<string> vars;
  for (int i=0;i<varTs->size();i=i+1)
      vars.insert(vars.end(),((*varTs)[i])->name);

  Expr* expr = p->snd;

  delete varTs;
  delete p->fst->fst;
  delete p->fst;
  delete p;

  return new FunDef(name->name,vars,expr);
}

ConstrDef* mkConstr(Pair<Token*,Number*>* p){
 string name = p->fst->name;
 int num = p->snd->val;
 delete p;
 return new ConstrDef(name,num);}

Fab4* mkFab4(vector<ConstrOrFun*>* constrOrFuns){
  vector<ConstrDef*> constrDefs;
  vector<FunDef*> funDefs;

  for (int i=0;i< constrOrFuns->size();i++) {
    if ((*constrOrFuns)[i] ->isLeft())
      constrDefs.insert
       (constrDefs.end()
       ,((Left<ConstrDef*,FunDef*>*)((*constrOrFuns)[i]))->left); 
    else
      funDefs.insert
       (funDefs.end()
       ,((Right<ConstrDef*,FunDef*>*)((*constrOrFuns)[i]))->right); 
    delete (*constrOrFuns)[i];
  }
  delete constrOrFuns;
  return new Fab4(constrDefs,funDefs,"GM Code");
}
}


