#ifndef __FAB4_TOKEN_H
#define __FAB4_TOKEN_H ;

#include <string>

using namespace std;

namespace fab4 {

class Token{
 public:
  string name;
  Token(string name);
  virtual bool eq(Token* t);
  virtual bool isCName();
  virtual bool isName();
  virtual bool isNumber();
  virtual bool isCharLiteral();
  virtual bool isStringLiteral();
  virtual bool isOperator();
  virtual string toString();
};

class CName:public Token{
 public:
  CName(string name);
  virtual bool isCName();
  virtual bool eq(Token* t);
  virtual string toString();
};

class Name:public Token{
 public:
  Name(string name);
  virtual bool isName();
  virtual bool eq(Token* t);
  virtual string toString();
};

class Number:public Token{
 public:
  int val;
  Number(int val);
  virtual bool isNumber();
  virtual bool eq(Token* t);
  virtual string toString();
};

class StringLiteral:public Token{
 public:
  StringLiteral(string s);
  virtual bool eq(Token* t);
  virtual bool isStringLiteral();
};

class CharLiteral:public Token{
 public:
  char c;
  CharLiteral(char c);
  virtual bool eq(Token* t);
  virtual bool isCharLiteral();
};

class Operator:public Token{
  public:
   string op;
   Operator(string op);
   virtual bool isOperator();
   string toString();
};

static Token* ARROW_T     = new Token("->"); 
static Operator* EQ_T     = new Operator("="); 
static Operator* ADD_T    = new Operator("+"); 
static Operator* SUB_T    = new Operator("-"); 
static Operator* MULT_T   = new Operator("*"); 
static Operator* DIV_T    = new Operator("/"); 
static Operator* LT_T     = new Operator("<"); 
static Operator* GT_T     = new Operator(">"); 
static Operator* GE_T     = new Operator(">="); 
static Operator* LE_T     = new Operator("<="); 
static Token* CONSTR_T    = new Token("constr"); 
static Token* CASE_T      = new Token("case"); 
static Token* OF_T        = new Token("of"); 
static Token* LPAR_T      = new Token("("); 
static Token* RPAR_T      = new Token(")"); 
static Token* SEMICOLON_T = new Token(";"); 
static Number* NUMBER_T   = new Number(0); 
static StringLiteral* STRING_T   = new StringLiteral(""); 
static CharLiteral* CHAR_T   = new CharLiteral(' '); 
static Token* NAME_T      = new Name("NAME_T"); 
static Token* CNAME_T     = new CName("CNAME_T"); 

bool tokenEq(Token* x,Token* y);

}
#endif


