#include <map>
#include <vector>
#include <string>
#include <fstream>

#include "../name/panitz/util/Show.h"
#include "Fab4Parser.h"
#include "Instruction.h"

using namespace std;
using namespace name::panitz::util;

namespace fab4 {

class GenExprCode:public ExprVisitor{
  public:

    vector<Instruction*>* result;

    vector<string> functions;
    std::map<string,int> constructors;
    std::map<string,int> constructorArity;

    vector<Pair<string,int> > env;
   
    int lookUp(string n){
      for (int i=0;i<env.size();i++)
        if (env[i].fst==n) return env[i].snd;
      throw ("unknown variable: "+n);
    }

    GenExprCode
      (vector<Instruction*>* r
      ,vector<Pair<string,int> >& e
      ,vector<string>& fs
      ,std::map<string,int>& cs,std::map<string,int>&  arity){
        result=r;
        env=e;
        functions=fs;
        constructors=cs;
        constructorArity=arity;
    }

    virtual void eval(Expr* arg){
      throw ("unmatched visitor pattern for: "+show(arg));
    }

    virtual void eval(NumExpr* arg){

    }

    static bool stringEq(string s1,string s2){return s1==s2;}

    virtual void eval(Var* arg){

    }

    virtual void eval(App* arg){

    }

    virtual void eval(OpExpr* arg){

    }

    virtual void eval(CharExpr* arg){

    }

    virtual void eval(Constr* arg){

    }

    virtual void eval(CaseExpr* arg){


      arg->e1->visit(this);
      result->insert(result->end(),new Eval());
      vector<CaseAlt*>* alts = arg->alts;
      vector<int>* jumpPointToEnd= new vector<int>();
      vector<Instruction*>* code=new vector<Instruction*>();

      CaseJump* jump=new CaseJump();
      result->insert(result->end(),jump);
      int startAddress = 0;
      for (int i=0;i<alts->size();i++){
        CaseAlt* alt=(*alts)[i];
        (*jump)[constructors[alt->constr]]=startAddress;

        vector<Instruction*>* altCode=new vector<Instruction*>();
        int n= alt->vars.size();
        altCode->insert(altCode->end(),new Split(n));
        vector<Pair<string,int> > newEnv=env;
        for (int j=0;j<env.size();j++)
          newEnv[j].snd=newEnv[j].snd+n;
        for (int j=0;j<n;j++){
          newEnv.insert(newEnv.begin(),Pair<string,int>(alt->vars[j],j));
        }
        alt->alt->visit(new GenExprCode
         (altCode,newEnv,functions,constructors,constructorArity));
        altCode->insert(altCode->end(),new Slide(n));
        altCode->insert(altCode->end(),new Jump(0));
        code->insert(code->end(),altCode->begin(),altCode->end());;
        startAddress=code->size(); 
        jumpPointToEnd->insert(jumpPointToEnd->end(),startAddress-1);
     }
     for (int i=0;i<jumpPointToEnd->size();i++){
      int jumpPoint=(*jumpPointToEnd)[i];
      (*code)[jumpPoint]=new Jump(startAddress-jumpPoint-1);
     }
     result->insert(result->end(),code->begin(),code->end());;


    }

};

vector<Instruction*>* genCode 
  (FunDef* fun
  ,vector<Instruction*>* result
  ,vector<string> fs
  ,std::map<string,int>& constr
  ,std::map<string,int>& ar){    

    std::vector<Pair<string,int> > env;
    for (int i=0;i<fun->args.size();i++)
      env.insert(env.end(),Pair<string,int>(fun->args[i],i));
    GenExprCode* genCoder = new GenExprCode(result,env,fs,constr,ar);
    fun->expr->visit(genCoder);
    if(fun->args.size()>0){
      result->insert(result->end(),new Update(fun->args.size()));
      result->insert(result->end(),new Pop(fun->args.size()));
    }
    result->insert(result->end(),new Unwind());
    
    return result;
};

vector<Instruction*>* genCode (Fab4* fab4){    
  vector<string> fs;
  vector<Instruction*>* result=new vector<Instruction*>();
  std::map<string,int> constr;
  std::map<string,int> arity;

  for (int i=0;i<fab4->constructors.size();i++){
    constr[fab4->constructors[i]->name]=i;
    arity[fab4->constructors[i]->name]=fab4->constructors[i]->arity;
  }

  for (int i=0;i<fab4->functions.size();i++)
    fs.insert(fs.end(),fab4->functions[i]->name);
  for (int i=0;i<fab4->functions.size();i++)  
    genCode(fab4->functions[i],result,fs,constr,arity);
  return result;
}
}

using namespace fab4;

int main(int argc,char* args[]){
 try {
  string a1 = args[1];
  ParsResult<Fab4*,Token>* res = parsFab4(a1+".fab4");
  string a1gmc=a1+".gmc";
  char fi [a1gmc.size()];
  fi [a1gmc.size()]='\0';
  a1gmc.copy(fi,a1gmc.size());
  ofstream out(fi,ios::out);

  out<<FA<<B4;
     
  cout << show(res->getFurtherToken())<<endl;

  Fab4* fab4=res->getResult();
  writeString(fab4->name,out);

  std::map<string,int> constr;
  std::map<string,int> arity;
  for (int i=0;i<fab4->constructors.size();i++){
    constr[fab4->constructors[i]->name]=i;
    arity[fab4->constructors[i]->name]=fab4->constructors[i]->arity;
  }

  vector<string> fs;
  vector<Instruction*>* code=new vector<Instruction*>();

  for (int i=0;i<fab4->functions.size();i++)
    fs.insert(fs.end(),fab4->functions[i]->name);

  // insert standard functions into machine
  fs.insert(fs.end(),"staticJavaCall");
  fs.insert(fs.end(),"javaCall");
  fs.insert(fs.end(),"getRuntime");  

  writeInt(fs.size(),out);

  writeString("staticJavaCall",out);
  writeInt(0,out);
  writeInt(2,out);
  code->insert(code->end(),new Push(1));
  code->insert(code->end(),new Eval());
  code->insert(code->end(),new Push(1));
  code->insert(code->end(),new Eval());
  code->insert(code->end(),new StaticJavaCall());
  code->insert(code->end(),new Unwind());

  writeString("javaCall",out);
  writeInt(code->size(),out);
  writeInt(3,out);
  code->insert(code->end(),new Push(2));
  code->insert(code->end(),new Eval());
  code->insert(code->end(),new Push(2));
  code->insert(code->end(),new Eval());
  code->insert(code->end(),new Push(2));
  code->insert(code->end(),new Eval());
  code->insert(code->end(),new JavaCall());
  code->insert(code->end(),new Unwind());

  writeString("getRuntime",out);
  writeInt(code->size(),out);
  writeInt(0,out);
  code->insert(code->end(),new GetRuntime());
  code->insert(code->end(),new Unwind());

  // write function table and generate code
  for (int i=0;i<fab4->functions.size();i++){  
    writeString(fab4->functions[i]->name,out);
    writeInt(code->size(),out);
    writeInt(fab4->functions[i]->args.size(),out);

    genCode(fab4->functions[i],code,fs,constr,arity);
  }

  // write byte code for constructor table
  writeInt(fab4->constructors.size(),out);
  for (int i =0;i<fab4->constructors.size();i++)
    writeString(fab4->constructors[i]->name,out);
  
  //write byte code for functions
  writeInt(code->size(),out);
  for (int i=0;i<code->size();i++){
    ((*code)[i])->write(out);
  }

  out<<FA<<B4<<flush;
  cout<<show(code)<<endl;
 }catch (string s){
  cout << "error: "<<s<<endl;
 }
}


