#include <map>
#include <vector>
#include <string>
#include <fstream>

#include "../name/panitz/util/Show.h"
#include "Fab4Parser.h"
#include "Instruction.h"

using namespace std;
using namespace name::panitz::util;

namespace fab4 {

void writeInt(int i,ostream& s){
   byte b1 = i/256/256/256%256;
   byte b2 = i/256/256%256;
   byte b3 = i/256%256;
   byte b4 = (byte)(i%256);
   s<<b1;
   s<<b2;
   s<<b3;
   s<<b4;
}

void writeString(string str,ostream& s){
  writeInt(str.size(),s);
  char cs [str.size()];
  
  for (int i=0;i<str.size();i++){
      s<<'\0';
      s<<str[i];
  };
}

string End::toString(){return "END";}
void End::write(ostream& out){out<<END;}

Jump::Jump(int n):n(n){}
string Jump::toString(){return "JUMP "+show(n);}
void Jump::write(ostream& out){out<<JUMP;writeInt(n,out);}

string Unwind::toString(){return "UNWIND";}
void Unwind::write(ostream& out){out<<UNWIND;}

string Eval::toString(){return "EVAL";}
void Eval::write(ostream& out){out<<EVAL;}

string Print::toString(){return "PRINT";}
void Print::write(ostream& out){out<<PRINT;}

string PrintString::toString(){return "PRINTSTRING";}
void PrintString::write(ostream& out){out<<PRINTSTRING;}

Output::Output(string s){this->s=s;}
string Output::toString(){return "OUTPUT "+s;}
void Output::write(ostream& out){out<<OUTPUT;writeString(s,out);}

string MkAp::toString(){return "MKAP";}
void MkAp::write(ostream& out){out<<MKAP;}

string JavaCall::toString(){return "JAVACALL";}
void JavaCall::write(ostream& out){out<<JAVACALL;}

string GetRuntime::toString(){return "GETRUNTIME";}
void GetRuntime::write(ostream& out){out<<GETRUNTIME;}

string StaticJavaCall::toString(){return "STATICJAVACALL";}
void StaticJavaCall::write(ostream& out){out<<STATICJAVACALL;}

string Add::toString(){return "ADD";}
void Add::write(ostream& out){out<<ADD;}

string Sub::toString(){return "SUB";}
void Sub::write(ostream& out){out<<SUB;}

string Mult::toString(){return "MULT";}
void Mult::write(ostream& out){out<<MULT;}

string Div::toString(){return "DIV";}
void Div::write(ostream& out){out<<DIV;}

string Le::toString(){return "LE";}
void Le::write(ostream& out){out<<LE;}

string Ge::toString(){return "GE";}
void Ge::write(ostream& out){out<<GE;}

string Lt::toString(){return "LT";}
void Lt::write(ostream& out){out<<LT;}

string Gt::toString(){return "GT";}
void Gt::write(ostream& out){out<<GT;}

string Eq::toString(){return "EQ";}
void Eq::write(ostream& out){out<<EQ;}

PushGlobal::PushGlobal(string n){name=n;}
string PushGlobal::toString(){return "PUSHGLOBAL "+name;}
void PushGlobal::write(ostream& out){
  out<<PUSHGLOBAL;writeString(name,out);
}

Push::Push(int n){this->n=n;}
string Push::toString(){return "PUSH "+show(n);}
void Push::write(ostream& out){out<<PUSH;writeInt(n,out);}

Pop::Pop(int n){this->n=n;}
string Pop::toString(){return "POP "+show(n);}
void Pop::write(ostream& out){out<<POP;writeInt(n,out);}

PushInt::PushInt(int n){this->n=n;}
string PushInt::toString(){return "PUSHINT "+show(n);}
void PushInt::write(ostream& out){out<<PUSHINT;writeInt(n,out);}

PushChar::PushChar(char n){this->n=n;}
string PushChar::toString(){return "PUSHCHAR "+show(n);}
void PushChar::write(ostream& out){out<<PUSHCHAR<<'\0'<<n;}

Slide::Slide(int n){this->n=n;}
string Slide::toString(){return "SLIDE "+show(n);}
void Slide::write(ostream& out){out<<SLIDE;writeInt(n,out);}

Update::Update(int n){this->n=n;}
string Update::toString(){return "UPDATE "+show(n);}
void Update::write(ostream& out){out<<UPDATE;writeInt(n,out);}

Pack::Pack(int n,int c){name=n;argc=c;}
string Pack::toString(){return "PACK "+show(name)+" "+show(argc);}
void Pack::write(ostream& out){
  out<<PACK;writeInt(name,out);writeInt(argc,out);}

struct ltstr
{
  bool operator()(int i1,int i2) const
  {
      return i1<i2;
  }
};

string CaseJump::toString(){return "CASEJUMP []"; }
template<class Iterator>
static void writeJumps(Iterator begin,Iterator end,ostream& out){
   for (Iterator it=begin;it!=end;it++){
     writeInt((*it).first,out);
     writeInt((*it).second,out);
   }
}
void CaseJump::write(ostream& out){
  out<<CASEJUMP; 
  writeInt(this->size(),out); 
  writeJumps(this->begin(),this->end(),out);}

Split::Split(int c){argc=c;}
string Split::toString(){return "SPLIT "+show(argc);}
void Split::write(ostream& out){out<<SPLIT;writeInt(argc,out);}

Instruction* getOp(Operator* opT){
  string op = opT->toString();
  if (op=="+")  return new Add();
  if (op=="-")  return new Sub();
  if (op=="*")  return new Mult();
  if (op=="/")  return new Div();
  if (op=="<")  return new Lt();
  if (op==">")  return new Gt();
  if (op=="<=") return new Le();
  if (op==">=") return new Ge();
  if (op=="=")  return new Eq();
  throw "unknown operator";
}
}

