#ifndef __INSTRUCTION_H
#define __INSTRUCTION_H 

#include <map>
#include <vector>
#include <string>
#include <fstream>

#include "../name/panitz/util/Show.h"
#include "Fab4Parser.h"
#include "Instruction.h"

using namespace std;
using namespace name::panitz::util;

namespace fab4 {
typedef char byte;

#define NOP          (byte)0
#define POP          (byte)1
#define END          (byte)2
#define JUMP         (byte)3
#define PUSH         (byte)4
#define UNWIND       (byte)5
#define EVAL         (byte)6
#define PRINT        (byte)7
#define PRINTSTRING  (byte)9
#define OUTPUT       (byte)10
#define MKAP         (byte)11
#define JAVACALL     (byte)12
#define PUSHGLOBAL   (byte)13
#define PUSHINT      (byte)14
#define PUSHCHAR     (byte)15
#define SLIDE        (byte)16
#define UPDATE       (byte)17
#define PACK         (byte)18
#define CASEJUMP     (byte)19
#define SPLIT        (byte)20
#define ADD          (byte)21
#define SUB          (byte)22
#define MULT         (byte)23
#define DIV          (byte)24
#define LE           (byte)25
#define GE           (byte)26
#define LT           (byte)27
#define GT           (byte)28
#define EQ           (byte)29
#define STATICJAVACALL (byte)30
#define GETRUNTIME   (byte)31

#define FA           (byte)250
#define B4           (byte)180

class  Instruction{
 public:
  virtual string toString()=0;
  virtual void write(ostream& out)=0;
};

class End:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Unwind:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Eval:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Print:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class PrintString:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class MkAp:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class JavaCall:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class GetRuntime:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class StaticJavaCall:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Add:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Sub:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Mult:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Div:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Le:public Instruction{
 public:  string toString();  virtual void write(ostream& out);};

class Ge:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Lt:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Gt:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Eq:public Instruction{
 public: string toString();  virtual void write(ostream& out);};

class Jump:public Instruction{
 public:
  int n;
  Jump(int n);
  string toString();  virtual void write(ostream& out);};

class Push:public Instruction{
 public:
  int n;
  Push(int n);
  string toString();  virtual void write(ostream& out);};

class Pop:public Instruction{
 public:
  int n;
  Pop(int n);
  string toString();  virtual void write(ostream& out);};

class PushInt:public Instruction{
 public:
  int n;
  PushInt(int n);
  string toString();  virtual void write(ostream& out);};

class Slide:public Instruction{
 public:
  int n;
  Slide(int n);
  string toString();  virtual void write(ostream& out);};

class Update:public Instruction{
 public:
  int n;
  Update(int n);
  string toString();  virtual void write(ostream& out);};

class Split:public Instruction{
 public:
  int argc;
  Split(int c);
  string toString();  virtual void write(ostream& out);};

class PushChar:public Instruction{
 public:
  char n;
  PushChar(char n);
  string toString();  virtual void write(ostream& out);};

class Pack:public Instruction{
 public:
  int name;
  int argc;
  Pack(int n,int c);
  string toString();  virtual void write(ostream& out);};

class Output:public Instruction{
 public:
  string s;
  Output(string s);
  string toString();  virtual void write(ostream& out);};

class PushGlobal:public Instruction{
 public:
  string name;
  PushGlobal(string n);
  string toString();  virtual void write(ostream& out);};

class CaseJump:  public std::map<int,int>
                ,public Instruction {
 public:
  string toString();  virtual void write(ostream& out);};

Instruction* getOp(Operator* opT);

template <typename InputIterator,typename Element>
bool contains(Element el,InputIterator begin,InputIterator end){
  for (InputIterator it = begin+1;it<end;it++){
    if (el == (*it)){ return true;}
  } 
  return false;
}
template <typename InputIterator,typename Element>
bool contains(Element el,InputIterator begin,InputIterator end
             ,bool(*eq)(Element,Element)){
  for (InputIterator it = begin+1;it<end;it++){
    if (eq(el,(*it))){return true;}
  } 
  return false;
}

void writeInt(int i,ostream& s);
void writeString(string str,ostream& s);
}
#endif


