package name.panitz.gm;
import java.io.*;
import java.util.*;
import name.panitz.util.*;

public class GmState{
  GmHeap heap;

  GC gc;

  GmStack stack=new GmStack();
  GmDump dump=new GmDump();

  GmGlobals globals;
  final String[] constructors;

  public final Instruction[] code;
  int pc=0;

  String name="GM Code";
  Writer output=new PrintWriter(System.out);

  public int alloc(Node n){return heap.alloc(n);}

  public boolean finalState(){return pc<0||pc>=code.length;}

  public void evalStep()throws Exception{
    Instruction i = code[pc];
    pc=pc+1;
    i.process(this);
  }

  public void eval()throws Exception{
    while (!finalState()) evalStep();
  }

  public String toString(){
    StringBuffer result=new StringBuffer("Stack:\n[");
    boolean first=true;

    for (Integer i:stack){
      if (first) first=false;else result.append(",");
      result.append(i+" -> "+heap.getNode(i).toString());
    }
    result.append("]");
    result.append("\n\nHeap\n");

    result.append(heap.toString());

    result.append("\n\nCode:\n[");
    for (Instruction c:code) result.append(c+";\n");
    result.append("]");

    return result.toString();
  }

  public GmState(Instruction[] code,GmHeap heap
                ,GmGlobals glob,String[] constructors){
    this.code=code;
    this.heap=heap;
    this.globals=glob;
    this.constructors=constructors;
    gc=GCFactory.getGCInstance(this);
  }

  public GmState(Instruction[] code){
    this(code,new GmHeap(),new GmGlobals(),new String[0]);
  }

  String[] progArgs;
  public String[] getProgArgs(){return progArgs;}

  Integer javaCallAddress=null;
  int getJavaCallAddress(){
    if (javaCallAddress==null)
      javaCallAddress
        =((NGlobal)heap.get((globals.get("javaCall")))).i;
    return  javaCallAddress;
  }

  Integer staticJavaCallAddress=null;
  int getStaticJavaCallAddress(){
    if (staticJavaCallAddress==null)
      staticJavaCallAddress
	=((NGlobal)heap.get((globals.get("staticJavaCall")))).i;
    return  staticJavaCallAddress;
  }

  public String getCurrentFunction(){
    final Map<Integer,String> functionEntries
       = new HashMap<Integer,String>();
    for (Map.Entry<String,Integer> entry:globals.entrySet()){
      final int codeAdr =((NGlobal)heap.get(entry.getValue())).i;
      functionEntries.put(codeAdr,entry.getKey());
    }
    Object [] values = functionEntries.keySet().toArray();
    Arrays.sort(values);

    int jumper=0;

    for (Object val:values){
      if (((Integer)val)>=pc)break; 
      jumper=(Integer)val;
    }
    return functionEntries.get(jumper);
  }
}

