package name.panitz.gm;
import name.panitz.util.Pair;

import static name.panitz.gm.ConstructorConstants.*;
import java.io.*;

public class MarshallingNode implements NodeVisitor<Object>{
  GmState st;
  public MarshallingNode(GmState st){this.st=st;}
  public Object eval(Node n)throws Exception{
    throw new Exception("unmatched pattern: "+n);
  }
  public Object eval(NInd n)throws Exception{
    st.stack.pop();
    st.stack.push(n.adr);
    return st.heap.get(n.adr).visit(this);}

  public Object eval(NAp n)throws Exception{
    throw new Exception("unmatched pattern: "+n);
  }

  public Object eval(NGlobal n)throws Exception{
    throw new Exception("unmatched pattern: "+n);
  }
    
  public Object eval(NJavaObject n){
    return n.o;
  }
  public Object eval(NNum n){return new Integer(n.n);}
  public Object eval(NChar n){return new Character(n.c);}

  public Object eval(NConstr constr)throws Exception{
    Writer newOut=new StringWriter();
    Writer orgOut=st.output;
    st.output=newOut;
    st.stack.push(st.alloc(constr));
    constr.visit(new PrintStringNode(st));
    st.output=orgOut;
    return newOut.toString();
  }

  static public int demarshalling(Object o,GmState st){
    final int neededSpace= neededHeapSpace(o);
    return demarshallingAux(o,st);
  }

  static  int neededHeapSpace(Object o){
    if (o instanceof String)return 1+2*((String)o).length();
    if (o instanceof Object[]) {
      Object[] xs=(Object[])o;
      int result = xs.length+1;
      for(Object x:xs)result=result+neededHeapSpace(x); 
      return result;
    }
    if (o instanceof Pair)
      return 1+neededHeapSpace(((Pair)o).e1)
              +neededHeapSpace(((Pair)o).e2);
    return 1;
  }

  static private int demarshallingAux(Object o,GmState st){
    if (o instanceof Boolean){
      int [] args={};
      if ((Boolean)o) {return st.alloc(new NConstr(TRUE,args));}
      return st.alloc(new NConstr(FALSE,args));
    }
    if (o instanceof Integer) {
      return st.alloc(new NNum((Integer)o));}
    if (o instanceof Character){
      return st.alloc(new NChar((Character)o));}
    if (o instanceof Pair) {
      int [] args
        ={demarshallingAux(((Pair)o).e1,st)
         ,demarshallingAux(((Pair)o).e2,st)};
      return st.alloc(new NConstr(PAIR,args));
   
    }
    if (o instanceof Object[]) {
      Object[] xs=(Object[])o;
      int aTail = st.alloc(new NConstr(NIL));
      for (int i=xs.length-1;i>=0;i--){
        int aHead=demarshallingAux(xs[i],st);                
        final int[] args={aHead,aTail};

        aTail=st.alloc(new NConstr(CONS,args));
      }
      return aTail;
    }
    if (o instanceof String){
      String s=(String)o;
      final int[] nilargs={};
      int aTail = st.alloc(new NConstr(NIL,nilargs));
      for (int i=s.length()-1;i>=0;i--){
        int aHead=st.alloc(new NChar(s.charAt(i)));                
        
        final int[] args={aHead,aTail};

        aTail=st.alloc(new NConstr(CONS,args));
      }
      return aTail;
    }
    return st.alloc(new NJavaObject(o));
  }
}

