package name.panitz.gm;

import java.io.*;
import java.util.*;
import static name.panitz.gm.CodeConstants.*;

public class ReadGm{
  static String readString(DataInputStream in)throws Exception{
    final int size = in.readInt();
    String result="";
    for (int i=0;i<size;i++)result=result+in.readChar();
    return result;
  }

  static GmState readGMCode(DataInputStream in) throws Exception{
    return readGMCode(in,false);}

  static GmState readGMCodeLog(DataInputStream in) throws Exception{
    return readGMCode(in,true);}

  static GmState readGMCode(DataInputStream in,boolean log)
                                               throws Exception{
    final byte[]fab4= new byte[2];
    in.read(fab4);
    if (fab4[0]!=FA||fab4[1]!=B4) throw new Exception("no gm code");

    final String name =readString(in);

    final GmHeap heap=new GmHeap();  
    final GmGlobals glob=new GmGlobals();

    final int fs=in.readInt();
    for (int i=0;i<fs;i++){
      final String n = readString(in);
      final int pc   = in.readInt();
      final int argc = in.readInt();
      int adr = heap.alloc(new NGlobal(argc,pc));
      glob.put(n,adr);
    }

    final int cs=in.readInt();
    String [] constructors = new String[cs];
    for (int i=0;i<cs;i++){
      constructors[i]=readString(in);
    }

    final int codeLength=in.readInt()+5;

    Instruction [] code=new Instruction[codeLength];
    for (int i=0;i<codeLength-5;i++)code[i]=readInstruction(in);  
    code[codeLength-5]=new PushGlobal("main");
    code[codeLength-4]=new Eval();
    code[codeLength-3]=new Print();
    code[codeLength-2]=new Output("\n");
    code[codeLength-1]=new End();

    GmState result ;
    if (log) result = new GmStateLog(code,heap,glob,constructors);
    else result = new GmState(code,heap,glob,constructors);
    
    result.pc=codeLength-5;
    result.name=name;
    return result;
  }

  static Instruction readInstruction(DataInputStream in)
                                               throws Exception{

    final byte b=in.readByte();
    switch (b){
      case NOP:        return new Nop();

    }
    throw new Exception("Illegal GM Code: "+b);
  }
}

