package name.panitz.gm;

import java.io.*;
import java.util.*;
import static name.panitz.gm.CodeConstants.*;

public class WriteGm{

  static void writeGMCode(DataOutputStream out, GmState st)
                                    throws Exception{

    out.write(FA);out.write(B4);        //FAB4

    write(out,st.name);                  //module name

    out.writeInt(
         st.globals.entrySet().size()); //number of functions

    for (Map.Entry<String,Integer> entry
                    :st.globals.entrySet()){
      write(out,entry.getKey());        //function name
      final NGlobal glob = (NGlobal)st.heap.get(entry.getValue());
      out.writeInt(glob.i);             //code pointer of function
      out.writeInt(glob.argc);          //number of function args
    }

    out.writeInt(st.constructors.length);//number of constructors
    for (String constr:st.constructors)
      write(out,constr);                //constructor name

    out.writeInt(st.code.length);
    for (Instruction i:st.code) writeInstruction(out,i);

    out.write(FA);out.write(B4);        //FAB4
    out.close();
  }

  static private void write(DataOutputStream out,String s)
                                                throws Exception{
    out.writeInt(s.length());      //length of string
    out.writeChars(s);             //characters of string
  }

  static void writeInstruction(DataOutputStream out,Instruction i)
                                                throws Exception{
   if (i instanceof End){out.write(END);
   }else if (i instanceof Unwind){out.write(UNWIND);  
   }else if (i instanceof Eval){out.write(EVAL);    
   }else if (i instanceof Print){out.write(PRINT);   
   }else if (i instanceof PrintString){out.write(PRINTSTRING);
   }else if (i instanceof MkAp){out.write(MKAP);    
   }else if (i instanceof JavaCall){out.write(JAVACALL);    
   }else if (i instanceof StaticJavaCall){out.write(STATICJAVACALL);    
   }else if (i instanceof GetRuntime){out.write(GETRUNTIME);    
   }else if (i instanceof Add){out.write(ADD);      
   }else if (i instanceof Sub){out.write(SUB);      
   }else if (i instanceof Mult){out.write(MULT);      
   }else if (i instanceof Div){out.write(DIV);      
   }else if (i instanceof Lt){out.write(LT);      
   }else if (i instanceof Gt){out.write(GT);      
   }else if (i instanceof Le){out.write(LE);      
   }else if (i instanceof Ge){out.write(GE);      
   }else if (i instanceof Eq){out.write(EQ);

   }else if (i instanceof Push){
     out.write(PUSH);out.writeInt(((Push)i).n);
   }else if (i instanceof Pop){
     out.write(POP);out.writeInt(((Pop)i).n);
   }else if (i instanceof Jump){
     out.write(JUMP);out.writeInt(((Jump)i).n); 
   }else if (i instanceof PushInt){
     out.write(PUSHINT);out.writeInt(((PushInt)i).n);
   }else if (i instanceof Slide){
     out.write(SLIDE);out.writeInt(((Slide)i).n);
   }else if (i instanceof Update){
     out.write(UPDATE);out.writeInt(((Update)i).n);
   }else if (i instanceof Split){
     out.write(SPLIT);out.writeInt(((Split)i).argc);

   }else if (i instanceof Output){
     out.write(OUTPUT);
     final String s = ((Output)i).s;
     out.writeInt(s.length());
     out.writeChars(s);
   }else if (i instanceof PushGlobal){out.write(PUSHGLOBAL);
     final String s = ((PushGlobal)i).name;
     out.writeInt(s.length());
     out.writeChars(s);
   }else if (i instanceof PushChar){out.write(PUSHCHAR); 
     final char[] c={((PushChar)i).n};
     out.writeChars(new String(c)); 
   }else if (i instanceof Pack){out.write(PACK);        
     final Pack p=(Pack)i;
     out.writeInt(p.name);
     out.writeInt(p.argc);
   }else if (i instanceof CaseJump){out.write(CASEJUMP); 
     final CaseJump jumps = (CaseJump)i;
     out.writeInt(jumps.entrySet().size());
     for (Map.Entry<Integer,Integer> entry:jumps.entrySet()){
       out.writeInt(entry.getKey());
       out.writeInt(entry.getValue());
     }
   }
  }
}

